/*\
|*| --------------------------------------------------------------------------
|*|	S3 Vision964 chipset driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|
|*|	This code was derived from the following source of information:
|*|	[1]	S3 Vision964 Data Book, (c) 1994 S3 Incorporated
\*/

/*\
|*|	Kernel includes
\*/

/*
#include <linux/mm.h>
#include <linux/mman.h>
*/

#include <asm/io.h>

extern void request_region(unsigned int from, unsigned int num,
                           const char *name);
extern int  check_region  (unsigned int from, unsigned int num);
extern void release_region(unsigned int from, unsigned int num);

#define DEBUG_IO

/*\
|*|     GGI includes
\*/

#undef DEBUG_IO
#define DRIVER_VERSION	"v0.00 ALPHA"

#include "graphdev.h"
#include "kgi.h"
#include "S3/Vision964.h"

#include "s3_io.inc"


static struct kgi_chipset
chipinfo =
{
	"S3 Incorporated",	/* manufacturer		*/
	"Vision 964",		/* model		*/
	GM_TEXT | GM_4BIT  |	/* supported modes	*/
	GM_8BIT | GM_15BIT |
	GM_16BIT| GM_24BIT,
	0,
	{ 2048, 2048 },		/* maximum resolution	*/

	0,			/* RAM size, to be initialised */
	{16000000, 75000000}	/* DCLK limits		*/
};

static struct statetable
default_state =
{
    0x23, 0x00,					/* MISC, FCTRL */
  {
     0x03, 0x00, 0x0F, 0x00, 0x0E			/* SEQ */
  },
  {
     0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT */
     0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE3,		/* 0x1x */
     0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x2x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x09, 0x00, 0x20, 0x10, 0x00, 0x00, 0x00,		/* 0x3x */
     0x4A, 0xAD, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x4x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,		/* 0x5x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x6x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x7x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  },
  {
     0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F,	/* GRC */
     0xFF
  },
  {
     0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC */
     0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
     0x8C, 0x00, 0x0F, 0x00, 0x00
  }
};

static int
detect_Vision964(void)
{
  #define COPYDEFAULT(X) (default_state.CRT[X]=inbCRT(X))
  register unsigned char foo;

  foo = COPYDEFAULT(0x30);
  if((foo & CR30_CHIPID) != CR30_Vision964ID)
    {
      printk("  Vision964 not detected (ID = %.2x).\n", foo & CR30_CHIPID);
      return -ECHIPSET;
    }
  TRACE(printk("  Vision964 rev %i detected.\n", foo & CR30_REVISION));


  foo = COPYDEFAULT(0x36);
   switch(foo & CR36_MEMSIZE_MASK)
    {
      case CR36_1MB: chipinfo.ramsize = 1;	break;
      case CR36_2MB: chipinfo.ramsize = 2;	break;
      case CR36_3MB: chipinfo.ramsize = 3;	break;
      case CR36_4MB: chipinfo.ramsize = 4;	break;
      case CR36_6MB: chipinfo.ramsize = 6;	break;
      case CR36_8MB: chipinfo.ramsize = 8;	break;
      default:  chipinfo.ramsize = 1;
                printk("%s: can't determine VRAM size, assume 1MB.\n",__FILE__);
    }
   #ifdef DEBUG
    switch(foo & CR36_RAMMASK)
     {
       case CR36_FASTPAGERAM: printk("  %i MB FP RAM, ",  chipinfo.ramsize); break;
       case CR36_EDO_RAM:     printk("  %i MB EDO RAM, ", chipinfo.ramsize); break;
       default:		      printk("  %i MB unknown RAM, ", chipinfo.ramsize);
     }
    switch(foo & CR36_BUSMASK)
     {
       case CR36_PCI      :   printk("PCI bus\n");        break;
       case CR36_VESALOCAL:   printk("Vesa Local bus\n"); break;
       default:               printk("unknown bus\n");
     }
   #endif
   chipinfo.ramsize *= 0x100000; /* convert MB to Byte */

  COPYDEFAULT(CRT_CONFIG2);
  default_state.CRT[0x3A] |= (inbCRT(0x3A) & (CR3A_PCIMASK | CR3A_REFRESHMASK));
  default_state.CRT[0x40] |= (inbCRT(0x40) & (CR40_BUSMASK));
  default_state.CRT[0x53] |= (inbCRT(0x53) & (CR53_PAR_SID)); 
  default_state.CRT[0x58] |= (inbCRT(0x58) & (CR58_SAM_256 | CR58_VL_T1_LATCH));
/*  default_state.CRT[0x5C] |= (inbCRT(0x5C) & (CR5C_GOUT_MASK)); */
  COPYDEFAULT(CRT_MEMCONTROL3);
  COPYDEFAULT(CRT_MEMCONTROL4);
  COPYDEFAULT(CRT_MEMCONTROL5);
  default_state.CRT[0x65] |= (inbCRT(0x65) & (CR65_SAM_MASK)); 
  default_state.CRT[0x66] |= (inbCRT(0x66) & (CR66_PCI_DISCONNECT));
  COPYDEFAULT(CRT_CONFIG3);

  /* temporary, using 64 KB LAW at 000a0000 */
  default_state.CRT[0x59] = 0x00;
  default_state.CRT[0x5A] = 0x0a;
/*  default_state.CRT[0x58] |= CR58_LAW_ENABLE; */

  return EOK;
  #undef COPYDEFAULT
}

/*\
|*|	GGI interface
\*/

int
kgi_chipset_init(void)
{

  TRACE(printk("S3 964 Chipset Driver "DRIVER_VERSION"\n"));

  outb(S3_DISABLE, S3_SUBSYSTEM);	/* chip wakeup		*/
  outb(S3_WAKEUP,  S3_SETUP);
  outb(S3_ENABLE,  S3_SUBSYSTEM);

  outbMISC(inb(MISCr));			/* adjust IO base	*/

  s3_unlock();

  if(! detect_Vision964())
    {
      register unsigned char foo;
      
      /* select CR42 to drive the clock select lines. */
      foo = inb(MISCr);
      if((foo & MISC_CLOCK_MASK) != MISC_EXT_CLOCK)
        { 
          outbCRT((foo &  MISC_CLOCK_MASK) >> 2, 0x42);
          outbMISC((foo & ~MISC_CLOCK_MASK) | MISC_EXT_CLOCK);
        }

      if(! check_region(S3IO_BASE, S3IO_REGION))
        { 
          request_region(S3IO_BASE, S3IO_REGION, "Vision964");
        } 
       else   
        {
	  printk("%s: warning: failed to allocate IO region.\n",__FILE__);
        }
      return EOK;
    }
   else
    return -ECHIPSET;
}

void
kgi_chipset_done(void)
{
  release_region(S3IO_BASE, S3IO_REGION);
  TRACE(printk("S3 Chipset Driver removed\n"));
  s3_lock();
}

struct kgi_chipset 
*kgi_chipset_get_info (void)
{
  return &chipinfo;
}

static void 
adjust_timing(struct kgi_mode *mode)
{
  register int foo = (mode->request.graphtype == GT_TEXT) 
  			   ? mode->request.textgrid.x : 8;

  #define ADJUST(X) mode->x.X -= (mode->x.X %foo)

  ADJUST(width);
  ADJUST(blankstart);
  ADJUST(syncstart);
  ADJUST(syncend);
  ADJUST(blankend);
  ADJUST(total);

  #undef ADJUST
}


int
kgi_chipset_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int Bpp = 1;

  if (mode->flags & FL_MON_INTERLACE)
    {
      TRACE(printk("%s: interlaced modes not supported (yet).\n",__FILE__));
      return -ECHIPSET; 
    }

  switch(mode->request.graphtype)
   {
     case GT_32BIT:	Bpp +=4;	/* (4+2+1+1)/2 = 4 Bpp	*/
     case GT_15BIT:
     case GT_16BIT:	Bpp +=2;	/*  (2+1+1)/2 =   2 Bpp	*/
     case GT_8BIT:	Bpp +=1;	/*    (1+1)/2 =   1 Bpp	*/
     case GT_4BIT:			/*        1/2 = 0.5 Bpp	*/
     case GT_TEXT:	break;
     default:
	TRACE(printk("%s: graphtype %i not supported.\n", __FILE__, \
		     mode->request.graphtype));
	return -ECHIPSET;
   }

  if (mode->lclk.mul != 1)
    return -ECHIPSET;

  switch (mode->lclk.div)
   {
     case 1: case 2: case 4: case 8: case 16: case 32:
	break;
     default:
	TRACE(printk("%s: invalid CLK to LCLK divider.\n",__FILE__));
	return -ECHIPSET;
   }

  #define XVIRTUAL (mode->request.virtual.x)
  #define YVIRTUAL (mode->request.virtual.y)

  switch (cmd)
   {
     case CMD_PROPOSE:

	if (mode->request.graphtype == GT_TEXT)
	  {
	    mode->magnify.x = mode->magnify.y = 1;
	    if (   (mode->request.textgrid.x < 7) 
	        || (9 < mode->request.textgrid.x))
	      {
	        TRACE(printk("%s: invalid char width (%i)\n",__FILE__,
		      mode->request.textgrid.x));
		return -EINVAL;
	      }

	    if ((2 * (XVIRTUAL / mode->request.textgrid.x)
		   * (YVIRTUAL / mode->request.textgrid.y)) > 0x20000 )
	      {
		TRACE(printk("%s: not enough text memory.\n",__FILE__));
		return -ENOMEM;
	      }
	    
	    mode->request.visible.x -= 
		(mode->request.visible.x % mode->request.textgrid.x);
	    mode->request.visible.y -= 
		(mode->request.visible.y % mode->request.textgrid.y);
	  }
         else	/* mode->request.graphtype == GT_TEXT */
	  {
	    mode->magnify.x = mode->magnify.y = 1;

	    if (XVIRTUAL <=  640) XVIRTUAL =  640; else
	    if (XVIRTUAL <=  800) XVIRTUAL =  800; else
	    if (XVIRTUAL <= 1024) XVIRTUAL = 1024; else
	    if (XVIRTUAL <= 1152) XVIRTUAL = 1152; else
	    if (XVIRTUAL <= 1280) XVIRTUAL = 1280; else
	    if (XVIRTUAL <= 1600) XVIRTUAL = 1600; else
	    if (XVIRTUAL <= 2048) XVIRTUAL = 2048;
              else
	       {
	         TRACE(printk("%s: %i horizontal pixels are too many.\n",\
			      __FILE__, XVIRTUAL));
	         return -ECHIPSET;
	       }
	
	    if ((XVIRTUAL* YVIRTUAL* Bpp / 2) > chipinfo.ramsize)
	      {
	        TRACE(printk(\
		"%s: not enough memory to do virtual %ix%i at %i bpp",\
		__FILE__, XVIRTUAL, YVIRTUAL, Bpp*4));
	        return -ENOMEM;
	      }

	    mode->request.visible.x &= 0x7FFFFFF8;
	    mode->request.visible.y &= 0x7FFFFFF8;
	}

	mode->x.width = mode->request.visible.x;
	mode->y.width = mode->request.visible.y;

	return EOK;

     case CMD_LOWER:

	if (mode->dclk < chipinfo.clock.min)
	  {
	    TRACE(printk("%s: DCLK too low (%i).\n",__FILE__, mode->dclk));
	    return -ECHIPSET;
          }

        if (mode->request.graphtype == GT_TEXT)
          mode->dclk = (mode->dclk > 40000000) ? 40000000 : mode->dclk;
         else 
	  mode->dclk = (mode->dclk > chipinfo.clock.max)
	 		  ?  chipinfo.clock.max : mode->dclk;

	adjust_timing(mode);
	return EOK;

     case CMD_RAISE:

	if ((mode->dclk > chipinfo.clock.max) || 
	    (mode->dclk > 40000000  &&  mode->request.graphtype == GT_TEXT))
	  {
	    TRACE(printk("%s: DCLK too high (%i).\n",__FILE__, mode->dclk));
	    return -ECHIPSET;
          }

	mode->dclk = (mode->dclk < chipinfo.clock.min)
			?  chipinfo.clock.min : mode->dclk;

	adjust_timing(mode);
	return EOK;

     case CMD_CHECK:
	if (mode->request.graphtype == GT_TEXT)
	  {
	     if (   (mode->magnify.x != 1) || (mode->magnify.y != 1)
		 || (mode->request.textgrid.x < 7) 
		 || (9 < mode->request.textgrid.x)
		 || (mode->request.textgrid.y < 1)
		 || (mode->request.textgrid.y > 16)
		 || ((2 * (XVIRTUAL / mode->request.textgrid.x)
			* (YVIRTUAL / mode->request.textgrid.y)) > 0x20000)
		 || (mode->dclk > 40000000) 
		 || (mode->dclk < chipinfo.clock.min))
		return -ECHIPSET;
	  }
	 else
	  {
	    switch (XVIRTUAL)	
	     {
	       case 640:  case 800:  case 1024:	case 1152:
	       case 1280: case 1600: case 2048:
	     	  break;
	       default:
	  	  return -ECHIPSET;
	     }

  	    if (   ((XVIRTUAL * YVIRTUAL * Bpp / 2) > chipinfo.ramsize)
	        || (mode->magnify.x != 1) || (mode->magnify.y != 1)	/* !!! temp */
	        || (mode->clk.mul != 1)   || (mode->clk.div != 1)
	        || (mode->dclk < chipinfo.clock.min) 
	        || (chipinfo.clock.max < mode->dclk))
	      return -ECHIPSET;
	  }

	#define TM(X) (   ((unsigned) chipinfo.maxdots.x < mode->x.X)\
		       || ((unsigned) chipinfo.maxdots.y < mode->y.X))
	if (   TM(width)   || TM(blankstart) || TM(syncstart)
	    || TM(syncend) || TM(blankend)   || TM(total))
	  return -ECHIPSET;
 	#undef TM

	adjust_timing(mode);
	return EOK;

     default:
	INTERNALERROR;
	return -ECHIPSET;
   }

  #undef XVIRTUAL 
  #undef YVIRTUAL
}

void
kgi_chipset_set_mode (struct kgi_mode *mode)
{
  struct statetable setup = default_state;
  struct kgi_ramdac *dac;
  register int foo;
  int DCLKdiv;
  int EC00 = EC00_IRQ_MASK;
  int EC01 = EC01_ENH_ENB | EC01_LAW_ENB;

  /*\
  |*|	Setup bits per pixel and logical screen width
  \*/

  switch(mode->request.graphtype)
   {
     case GT_TEXT:
	setup.CRT[0x17] =   CR17_WORDMODE 
			  | CR17_ENABLE_SYNC | CR17_CGA_BANKING 
			  | CR17_HGC_BANKING;
        
	foo = 2 * mode->request.virtual.x / mode->request.textgrid.x; 
        break;
     case GT_4BIT:
	setup.CRT[0x50] |= CR50_4BPP;
	EC01 |= EC01_4BPP;
	foo = mode->request.virtual.x >> 4;
	break;
     case GT_8BIT:
	setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
	setup.CRT[0x50] |= CR50_8BPP;
	foo = mode->request.virtual.x >> 3;
	break;
     case GT_15BIT:
     case GT_16BIT:
	setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
	setup.CRT[0x50] |= CR50_16BPP;
	setup.CRT[0x67] |= CR67_NO_INVERT_VCLK;
	foo = mode->request.virtual.x >> 2;
	break;
     case GT_32BIT:
	setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
	setup.CRT[0x50] |= CR50_32BPP;
	foo = mode->request.virtual.x >> 1;
	break;

     default:
	INTERNALERROR;
	return;
   }
   setup.CRT[CRT_LOGICALWIDTH]  |= foo & 0xFF;
   setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
   setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

   if(mode->request.graphtype == GT_TEXT)
     {
       setup.SEQ[0x04] = 0x02; /* !!! temporary */  
       setup.SEQ[0x02] = 0x03; /* write enable planes 0 and 1 */
       switch((DCLKdiv = mode->request.textgrid.x))
	{
	  case 7: setup.CRT[0x32] |= CR32_7DOT_CHCLK; break;
	  case 8: setup.SEQ[0x01] |= SR01_8DOT_CHCLK; break;
	  case 9: break;
	  default:
	     INTERNALERROR;
	     return;
	}
       setup.CRT[0x09] |= mode->request.textgrid.y;
       setup.ATC[0x13] = (mode->request.textgrid.x == 9) ? 0x08 : 0x00; 
       setup.CRT[0x31] &= 0xF7;

       setup.CRT[0x0A] = 0x00;
       setup.CRT[0x0B] = 0x08;
       setup.CRT[0x17] |= CR17_CGA_ADDR_WRAP;

       setup.GRC[0x05] = GR5_ODD_EVEN;

     /*  setup.GRC[0x01] = 0x03; */
       setup.GRC[0x06] = 0x0E;
       setup.GRC[0x07] = 0x00;

     }
    else /* mode->request.graphtype == GT_TEXT */
     {
       DCLKdiv = 8;

       setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
       setup.ATC[0x10] |= AR10_GRAPHICS;
       setup.GRC[0x06] |= GR6_GRAPHMODE;

       setup.CRT[0x09] |= /* !!! TM->ymagnify & CR09_CHARHEIGHT_MASK; */ 0;
       setup.CRT[0x09] |= (mode->flags & FL_CHIP_DOUBLESCAN) 
				? CR09_DOUBLESCAN : 0;

       switch(mode->request.virtual.x)
        { 
          case  640: setup.CRT[0x50] |= CR50_640; break;
          case  800: setup.CRT[0x50] |= CR50_800; break;
          case 1024: setup.CRT[0x50] |= CR50_1024; break;
          case 1152: setup.CRT[0x50] |= CR50_1152; break;
          case 1280: setup.CRT[0x50] |= CR50_1280; break;
          case 1600: setup.CRT[0x50] |= CR50_1600; break;
          case 2048: setup.CRT[0x50] |= CR50_2048;
		     setup.CRT[0x31] |= CR31_2PAGE_SCRN; 
		     break;
          default: INTERNALERROR;
        }

       /*\
       |*|  ???  I am not sure whether the following is right. May be these bits
       |*|  ???  have to be set according to the ammount of VRAM installed ?
       |*|  ???  If anyone has such a hardware, please contact me.
       \*/

       dac = kgi_ramdac_get_info();
       if(dac->dactype & DAC_VRAM)
         {
           setup.CRT[0x66] |= CR66_PIXELBUS_OFF;
           switch(dac->dactype & DAC_PB_MASK)
            { /* some are NOT_TESTED yet. If you have such a configuration
	       * please report your experience.  May be CR65 needs to be updated
	       * too ??? (FIXME) */
	      case DAC_16BIT_PB:
	 	setup.CRT[0x66] |= CR66_32S_SID;
	 	NOT_TESTED;
	 	break;
	      case DAC_32BIT_PB:
	 	setup.CRT[0x66] |= CR66_32P_SID;
/* ???	 	setup.CRT[0x53] |= CR53_PAR_SID; */
/* ???	 	setup.CRT[0x51] |= CR51_DISABLE_SPLIT;*/
	 	break;
	      case DAC_64BIT_PB:
	 	setup.CRT[0x66] |= CR66_64S_SID;
	 	NOT_TESTED;
		break; 
	      case DAC_128BIT_PB:
	 	setup.CRT[0x66] |= CR66_128S_SID;
	 	NOT_TESTED;
	 	break; 
	      default:
	 	printk("%s: Don't know how to handle DAC type %.8x.\n",\
	 		__FILE__,dac->dactype);
		return;
            }
         }
        else /* dac->dactype & DAC_VRAM */ 
         { /* strange: a 964 with a RAMDAC that doesn't support VRAM SCLK. */
           printk(\
          "%s: The selected RAMDAC doesn't support VRAM shift clock but this\n"\
          "%s: driver requires the DAC to use VRAM mode in graphics modes.\n"\
          "%s: This may be beause you selected an inappropriate RAMDAC for\n"\
          "%s: your card. If not, please report to linux-ggi@eskimo.com.\n",\
           __FILE__,__FILE__,__FILE__,__FILE__);
          return;
         }
     }

  /*\
  |*|	setup horizontal timing
  \*/
 
   foo = (mode->x.width / DCLKdiv) - 1;		/* horizontal display end */
     setup.CRT[0x01] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

   foo = (mode->x.blankstart / DCLKdiv) - 1;	/* horizontal blank start */
     setup.CRT[0x02] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

   foo = (mode->x.syncstart / DCLKdiv);		/* horizontal sync start  */
     setup.CRT[0x04] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

   foo = (mode->x.syncend / DCLKdiv);		/* horizontal sync end    */
     setup.CRT[0x05] |= foo & CR05_HSE_MASK;
     setup.CRT[0x5D] |= ((foo- (mode->x.syncstart/DCLKdiv)) > 0x20)
				 ? CR5D_HSYNCEND_B5 : 0;

   foo = (mode->x.blankend / DCLKdiv); 		/* horizontal blank end   */
     setup.CRT[0x03] |= foo & CR03_HBE_MASK;
     setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
     setup.CRT[0x5D] |= ((foo - (mode->x.blankstart/DCLKdiv)) > 0x40)
				 ? CR5D_HBLANKEND_B6 : 0;

   foo = (mode->x.total / DCLKdiv) - 5;		/* horizontal total       */
     setup.CRT[0x00] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;

  /*\
  |*|	setup vertical timing
  \*/
 
   foo = mode->y.width - 1;			/* vertical display end   */
     setup.CRT[0x12] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;

   foo = mode->y.blankstart;			/* vertical blank start   */
     setup.CRT[0x15] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
     setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;
    
   foo = mode->y.syncstart;			/* vertical sync start    */
     setup.CRT[0x10] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

   foo = mode->y.syncend;				/* vertical sync end      */
     setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;
 
   foo = mode->y.blankend + 1;			/* end vertical blank     */
     setup.CRT[0x16] |= foo & 0xFF;
  
   foo = mode->y.total - 2;			/* vertical total         */
     setup.CRT[0x06] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;
 
   setup.CRT[CRT_INTERLACESTART]|= mode->y.total >> 1; /* ??? wrong ? (FIXME) */
   if(mode->flags & FL_MON_INTERLACE)
    {
       printk("%s:%i: no interlaced modes (yet).\n",__FILE__,__LINE__);
       setup.CRT[0x42] |= CR42_INTERLACED;
       return; /* temporary */
    }

  setup.MISC |= mode->x.polarity ? MISC_NEG_HSYNC : 0;
  setup.MISC |= mode->y.polarity ? MISC_NEG_VSYNC : 0;

  for (foo = 0; !((1 << foo) & mode->lclk.div); foo++);
  setup.CRT[0x66] |= foo & CR66_VCLK_DIV_MASK;

  /*\
  |*|	Setup the FIFO control parameters.
  \*/

  foo = (mode->x.total + mode->x.syncstart)/2/DCLKdiv;
  setup.CRT[0x3B] |= foo & 0xFF;
  setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;

  /*\
  |*|	Ok. We'll do the programming now.
  \*/

/*  setup.CRT[0x3C] = 0x3D; */

#if 10
  outbCRT(inb(CRT_EXTSYNC1) | CR56_SYNC_OFF, CRT_EXTSYNC1);
  kgi_monitor_set_mode(mode);
  kgi_ramdac_set_mode(mode);
  kgi_clock_set_mode(mode);
  restore_state(&setup);
#endif

#if 0 /* this has to be done only for graphics mode !!! */
  outw_p(EC00 | EC00_RESET, ECR_MODECONTROL);
  outw_p(EC00, ECR_MODECONTROL);
  outw_p(EC01, ECR_FCTRL);
#endif


  dump_state(&setup);

  memset((void *) 0x000A0000, 0x12, 0x10000); 
  memset((void *) 0x000B0000, 0x34, 0x10000);
  memset((void *) 0x000B8000, 0x56, 0x08000);

#if 0
  outw(0x1000,0xbee8);	/* temporary */
  outw(0x2000,0xbee8);
  outw(0x3fff,0xbee8);
  outw(0x4fff,0xbee8);
#endif
}


int
kgi_chipset_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  return -ECHIPSET;
}

int
kgi_graphics_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		   unsigned long arg)
{
  return -ECHIPSET;
}

