/*\
|*| --------------------------------------------------------------------------
|*|     GGI device file definitions
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Andreas Beck, [becka@hp.rz.uni-duesseldorf.de]
|*|			  and Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	This program is distributed in the hope that it will be useful,
|*|	but WITHOUT ANY WARRANTY; without even the implied warranty of
|*|	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|*|	GNU General Public License for more details.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#ifndef _graphdev_h
#define _graphdev_h

#include <linux/fs.h>
#include <linux/ioctl.h>

#include "ggi.h"
#include "err.h"

/*\
|*|	Definitions for device-registration
\*/

#define	DEVICEFILE_MAJOR 8
#define DEVICEFILE_NAME  "graphics"
#define	DEVICEFILE_MAGIC ((GGI_VERSION << 8) | 0xAB)

/* ???	bit coding according to <linux/ioctl.h>. Note that this file may
 * ???	need slight patching:
 * ???	in line "#define IOCSIZE_MAX ..."  "IOC_SHIFT"  must be "IOCSIZE_SHIFT"
 */

/*\
|*|	The ioctl's
|*|
|*|	Bit coding is according to the rules in <linux/ioctl.h> now. The command
|*|	part is further divided with the upper 3 bits giving the subsystem and 
|*|	the remaining 13 bit the command. 
\*/

#define	SIZE(t)	((sizeof(t) << IOCSIZE_SHIFT) & IOCSIZE_MASK)
#define	IOC(c)  (IOC_VOID | c)

/*\ !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
|*| !!!	NOTE: DON NOT CHANGE WITHOUT CHANGING THE KERNEL DRIVER INTERFACE !!!
\*/ 

#define IOC_DRIVER		0x0000	/* kernel driver	*/
#define IOC_MONITOR		0x2000	/* monitor driver	*/
#define IOC_RAMDAC		0x4000	/* ramdac driver	*/
#define IOC_CLOCK		0x6000	/* clock driver		*/
#define IOC_CHIPSET		0x8000	/* chipset driver	*/
#define IOC_GRAPHICS		0xA000	/* graphic driver	*/
/*				0xC000	   not used yet		*/
/*				0xE000	   not used yet		*/

#define	IOC_SUBSYSTEM_MASK	0xE000
#define	IOC_SUBSYSTEM_SHIFT	13
#define	IOC_COMMAND_MASK	0x1FFF


/* ???	man koennte ja schon mit einem Auge uf das X11 Protokoll schielen, 
 * ???	irgendwann muss man es eh implementieren und dann waers ein bisserl
 * ???	leichter...
 */

/*\
|*|	ioctls for the graphics device
\*/

#define IDR(i,t,c)	(IOC_ ## i | SIZE(struct t) | IOC_DRIVER | c)

#define	IOC_DRIVER_TESTDRIVER	IOC(IOC_DRIVER |		0x0000)
#define	IOC_DRIVER_TESTMODE	IDR(INOUT, ggi_mode,		0x0001)

#define IOC_DRIVER_SETGRAPHMODE	IDR(INOUT, ggi_mode,		0x0002)
#define IOC_DRIVER_GETGRAPHMODE	IDR(OUT,   ggi_mode,		0x0003)
#define	IOC_DRIVER_GRAPHICS	IOC(IOC_DRIVER | 		0x0004)

#define	IOC_DRIVER_SETTEXTMODE	IDR(INOUT, ggi_mode,		0x0005)
#define	IOC_DRIVER_GETTEXTMODE	IDR(OUT,   ggi_mode,		0x0006)
#define	IOC_DRIVER_TEXT		IOC(IOC_DRIVER |		0x0007)


#define ICH(i,t,c)	(IOC_ ## i | SIZE(struct t) | IOC_CHIPSET | c)

#define IOC_CHIP_SETVISWIN	ICH(IN,   ggi_coord,		0x0000)
#define IOC_CHIP_GETVISWIN	ICH(OUT,  ggi_coord,		0x0001)
#define	IOC_CHIP_SETSPLITLINE	IOC(IOC_CHIPSET |		0x0002)
#define	IOC_CHIP_GETSPLITLINE	IOC(IOC_CHIPSET |		0x0003)

#define	IOC_CHIP_SETDEFFONT	ICH(IN,   ggi_bm_font,		0x0004)
#define	IOC_CHIP_GETDEFFONT	ICH(OUT,  ggi_bm_font,		0x0005)
#define	IOC_CHIP_SETFONTCHAR	ICH(IN,   ggi_bm_char,		0x0006)
#define	IOC_CHIP_GETFONTCHAR	ICH(OUT,  ggi_bm_char,		0x0007)

#define IGR(i,t,c)	(IOC_ ## i | SIZE(struct t) | IOC_GRAPHICS | c)

#define IOC_GRAPH_DRAWPIXEL	IGR(IN,    ggi_coord,		0x0000)
#define IOC_GRAPH_GETPIXEL	IGR(IN,	   ggi_coord,		0x0001)
#define IOC_GRAPH_PUTPIXEL	IGR(IN,    ggi_pixel,		0x0002)

						/* Optional from now on	*/
#define IOC_GRAPH_DRAWHLINE	IGR(IN,    ggi_line,		0x0008)
/*
   #define IOC_GRAPH_GETHLINE	IGR(INOUT, ,0x0009)
   #define IOC_GRAPH_PUTHLINE	IGR(INOUT, ,0x000a)
 */

#define IOC_GRAPH_DRAWVLINE	IGR(IN,	   ggi_rect,		0x0010)
/*
   #define IOC_GRAPH_GETVLINE	(IGR|0x0011)
   #define IOC_GRAPH_PUTVLINE	(IGR|0x0012)
 */

#define IOC_GRAPH_DRAWBOX	IGR(IN,    ggi_rect,		0x0020)
/*
   #define IOC_GRAPH_GETBOX	IGR(,,0x0021)
   #define IOC_GRAPH_PUTBOX	IGR(,,0x0022)
 */

#define IOC_GRAPH_FILLSCREEN	IOC(IOC_GRAPHICS | 0x0028)

#define	IOC_GRAPH_DRAWLINE	IGR(IN,	   ggi_rect,		0x0030)

#define IOC_GRAPH_DRAWELLIPSE	IGR(IN, ggi_ellipse, 0x0044)



#define IDAC(i,t,c)	(IOC_ ## i | SIZE(struct t) | IOC_RAMDAC | c)

#define IOC_RAMDAC_SETCLUT	IOC(IOC_RAMDAC |		0x0001)
#define IOC_RAMDAC_GETCLUT	IOC(IOC_RAMDAC |		0x0002)
/*
#define IOC_RAMDAC_SETHWCDATA	IDAC(,,0x0008)
#define IOC_RAMDAC_SETHWCPOS	IDAC(,,0x0009)
*/



#define IMON(i,t,c)	(IOC_ ## i | SIZE(struct t) | IOC_MONITOR | c)
/*
#define	IOC_MONITOR_GETINFO	IMON(0x0000)
#define IOC_MONITOR_SETINFO	IMON(0x0001)
#define IOC_MONITOR_SETPREFMODE	IMON(0x0010)
#define IOC_MONITOR_GETPREFMODE	IMON(0x0011)
#define IOC_MONITOR_CLEARPREF	IMON(0x0012)
#define IOC_MONITOR_POWERSAVE	IMON(0x0020)
*/

/* mmap-constants for the offset (which is actually ignored) */

/* ??? Unterschied zwischen LINEAR und FRAME ? Ich denke die Applikation soll
 * ??? nur noch einen linearen framebuffer 'sehen' ? 
 */

#define MMAP_LINEAR	0	/* Linear frame buffer */
#define MMAP_FRAME	1	/* banked frame buffer - optional ... for old apps */

/* MACROS : */

#if 0
/* ??? waere dann ueberfluessig. */

#define COPYPAR(size) \
	do { int err; int cnt = size; \
	    if ((err = verify_area(VERIFY_READ , (void *) arg, cnt))) \
	      return err; \
	     else \
	      memcpy_fromfs(parameters, (void *) arg, cnt); \
	   } \
	while(0)
#endif

#endif	/* #ifdef _graphdev_h */

