/*\
|*|	Generic Monitor driver for single frequency monitors
\*/

static struct kgi_timing
htiming =		/* horizontal timings in ns	*/
{
	,		/* width		*/ 
	,		/* blankstart = hactive	*/	
	,		/* syncstart		*/
	,		/* syncend		*/
	,		/* blankend		*/
	,		/* total		*/
	,		/* polarity		*/
};

static struct kgi_timing
vtiming =
{
	0,		/* width		*/ 
	,		/* blankstart = hactive	*/
	,		/* syncstart		*/
	,		/* syncend		*/
	,		/* blankend		*/
	,		/* total		*/
	0		/* polarity		*/
};

static struct kgi_monitor
monitor = {
		"Sony Corporation",	/* manufacturer		*/
		"CPD-1730",		/* model		*/
		GM_ALL,			/* supported modes	*/
		FL_MON_POWERSAVE |	/* special capabilities	*/
		FL_MON_INTERLACE, 
		{ 1024, 768 },		/* maximum resolution	*/

		{  382, 242 },		/* max picture size (mm)*/
		VGA | COLOR,		/* VGA color monitor	*/
		SYNC_NORMAL|POWER_SAVE,	/* normal syncing, VESA	*/
		{     0, 130000000 },	/* pixel clock range	*/
		{ 28000,     58000 },	/* hfreq range		*/
		{   55,        110 }	/* vfreq range		*/
	   };

/*\
|*|	GGI interface
\*/

int
kgi_monitor_init(void)
{
  if (monitor.hfreq.min != monitor.hfreq.max)
    {
      printk(" This is a fixed hfreq monitor driver but hfreq is %i-%i Hz!\n",\
	     monitor.hfreq.min, monitor.hfreq.max);
      return -EMONITOR;
    }

  TRACE(printk("%s %s Monitor Driver v"MON_VERSION"\n", \
        monitor.manufact, monitor.model));
  TRACE(printk("(%s - %s, %dx%d Pixel, %d mm x %d mm, %d-%d kHz, %d-%d Hz)\n",\
	  DISPLAYTYPE(monitor.disptype),\
	  (monitor.disptype & MONO) ? "monochrome" : "color",\
	  monitor.maxdots.x,\
	  monitor.maxdots.y,\
	  monitor.size.x,\
	  monitor.size.y,\
	  monitor.hfreq.min/1000 ,monitor.hfreq.max/1000,\
	  monitor.vfreq.min      ,monitor.vfreq.max ));
  return EOK;
}

void
kgi_monitor_done(void)
{
   TRACE(printk("%s %s Monitor driver removed.\n", \
         monitor.manufact, monitor.model));
}

struct kgi_monitor
*kgi_monitor_get_info (void)
{
  return &monitor;
}

int
kgi_monitor_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd *cmd)
{
  int hfreq, vfreq;
  long long foo;

  switch(*cmd)
   {
     case CMD_PROPOSE:
	mode->dclk = 
	*cmd = CMD_RAISE;
	return EOK;

     case CMD_RAISE:
	foo = mode->dclk * htiming.
	*cmd = CMD_CHECK;
	return EOK;

     case CMD_CHECK:

	if (   (mode->x.width > monitor.maxdots.x) 
	    || (mode->y.width > monitor.maxdots.y))
	  return -EMONITOR;

	hfreq = mode->dclk / mode->x.total;	/* hfreq near MONITOR_HFREQ ? */
	if (abs(MONITOR_HFREQ - hfreq) > MONITOR_HFREQ*3/100)
	  return -EMONITOR;

	vfreq = mode->dclk / (mode->x.total * mode->y.total);
        if ((vfreq < monitor.vfreq.min) || (mode->dclk < monitor.pfreq.min))
	  return -EMONITOR;

	if ((monitor.vfreq.max < vfreq) || (monitor.pfreq.max < mode->dclk))
	  return -EMONITOR;

	*cmd = CMD_READY;
	return EOK;

     default:
	INTERNALERROR;
	return -EMONITOR;
   }
}

int
kgi_monitor_set_mode(struct kgi_mode *TM)
{ 
   return EOK;
}

int
kgi_monitor_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  return -EMONITOR;
}
